#
# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

import os
import re
import shutil
import subprocess
import sys
import threading
import time
import tempfile
import json
from typing import Dict, List, Optional, Tuple
from concurrent.futures import ThreadPoolExecutor, as_completed
import datetime
from pathlib import Path

import psutil
from tabulate import tabulate
from tqdm.auto import tqdm

BASE_LOG_DIR = "/var/log/doca_installer_logs"
BFB_INSTALL_PATH = "/usr/sbin/bfb-install"
BF_LOG_DIR = None  # Will be set during setup_logging
DEFAULT_TIMEOUT = 120
CONSOLE_READ_TIMEOUT = 45
RSHIM_SERVICE_RESTART_TIMEOUT = 60
MAX_RETRIES = 5
BUDNLE_INSTALL_TIMEOUT = 2700  # 45 min
TIME_FOR_DEVICE_TO_BOOT = 2700  # 45 min
MLX_CONFIG_FULL_TIMEOUT = 300  # 5 min
TARGET_FW_VERSIONS = {}


class BFDevice:
    psid = None
    part_number = None
    serial_number = None
    mode = None
    rshim_device = None
    rshim_pci_address = None
    pre_install_nic_fw_version = None
    pre_install_atf_fw_version = None
    pre_install_uefi_fw_version = None
    pre_install_bmc_fw_version = None
    pre_install_cec_fw_version = None
    post_install_nic_fw_version = None
    post_install_atf_fw_version = None
    post_install_uefi_fw_version = None
    post_install_bmc_fw_version = None
    post_install_cec_fw_version = None
    running_dual_fw = None
    nic_fw_privileged_state = None
    mst_address = None
    required_power_cycle = False
    device_config_file = None
    upgrade_completed = False
    dpu_ready = False
    flint_dict = {}

    def __init__(self, bf_pci_address: str, bf_type: str):
        self.bf_pci_address = bf_pci_address
        self.bf_type = bf_type

    def _set_bf_mode(self):
        """
        Set DPU mode of operation [NIC/DPU]
        """
        mode = None
        mlxconfig_operation_mode_cmd = f"mlxconfig -d {self.bf_pci_address} -e q INTERNAL_CPU_OFFLOAD_ENGINE"
        mlxconfig_operation_mode_output, rc = run_command(mlxconfig_operation_mode_cmd)
        write_to_file(
            f"BF_{self.bf_pci_address.replace(':', '_')}_mlxconfig_check_nic_mode_cmd_output",
            mlxconfig_operation_mode_cmd,
            mlxconfig_operation_mode_output,
        )
        if rc != 0:
            mode = None
        for line in mlxconfig_operation_mode_output.splitlines():
            if "INTERNAL_CPU_OFFLOAD_ENGINE" in line:
                arr = line.split()
                if len(arr) > 4:
                    del arr[0]
                mode = "NIC" if "DISABLED" in arr[2] else "DPU"
        self.mode = mode

    def _set_nic_pn_sn(self):
        """
        Set Part number and Serial number of a BF DPU.
        """
        lspci_cmd = f"lspci -s {self.bf_pci_address} -vvv"
        lspci_output, rc = run_command(lspci_cmd)
        write_to_file(
            f"BF_{self.bf_pci_address.replace(':', '_')}_lspci_verbose_cmd_output",
            lspci_cmd,
            lspci_output,
        )
        part_number, serial_number = None, None
        if rc == 0:
            pn_match = re.search(r"\[PN\] Part number: (\S+)", lspci_output)
            sn_match = re.search(r"\[SN\] Serial number: (\S+)", lspci_output)
            part_number = pn_match.group(1) if pn_match else None
            serial_number = sn_match.group(1) if sn_match else None
        else:
            print(f"ERROR: Failed to get part number and serial number for BF device {self.bf_pci_address}")
        self.part_number, self.serial_number = part_number, serial_number

    def _set_privileged_mode(self):
        """
        Set the firmware privilege mode of a BF device
        """
        nic_fw_privileged_state = False
        mlxprivhost_cmd = f"mlxprivhost -d {self.bf_pci_address} q"
        mlxprivhost_output, rc = run_command(mlxprivhost_cmd)
        write_to_file(
            f"BF_{self.bf_pci_address.replace(':', '_')}_mlxprivhost_cmd_output",
            mlxprivhost_cmd,
            mlxprivhost_output,
        )
        if rc == 0:
            for line in mlxprivhost_output.splitlines():
                if "level" in line:
                    nic_fw_privileged_state = "PRIVILEGED" == line.split()[-1]
        else:
            print(f"ERROR: Failed to get privileged firmware state for BF device {self.bf_pci_address}")
        self.nic_fw_privileged_state = nic_fw_privileged_state

    def set_display_level(self, display_level="2"):
        """
        Set BF device rshim display level to giving value, or default value (2) if not specified.
        """
        display_level_command = f"echo 'DISPLAY_LEVEL {display_level}' > /dev/{self.rshim_device}/misc"
        display_level_output, rc = run_command(display_level_command)

    def _populate_flint_dict(self):
        """
        Run the flint command and populate the flint_dict field with parsed data.
        """
        flint_cmd = f"flint -d {self.bf_pci_address} q full"
        flint_output, rc = run_command(flint_cmd)
        write_to_file(
            f"BF_{self.bf_pci_address.replace(':', '_')}_flint_cmd_output",
            flint_cmd,
            flint_output,
        )

        if rc == 0:
            fw_details = {}
            current_key = None
            for line in flint_output.splitlines():
                if not line.strip():
                    continue
                if line.startswith(" ") or line.startswith("\t"):
                    if current_key is not None:
                        fw_details[current_key] += " " + line.strip()
                    continue
                if ":" in line:
                    key, value = line.split(":", 1)
                    current_key = key.strip()
                    fw_details[current_key] = value.strip()
            self.flint_dict = fw_details
        else:
            print(f"ERROR: Failed to run flint command ({flint_cmd}) for {self.bf_pci_address}")

    def _set_psid(self):
        """
        Set PSID using data from flint_dict.
        """
        self.psid = self.flint_dict.get("PSID")

    def _set_fw_version(self):
        """
        Set NIC firmware version using data from flint_dict.
        """
        running_fw_identifier = "FW Version(Running)"
        if running_fw_identifier in self.flint_dict:
            self.pre_install_nic_fw_version = self.flint_dict.get(running_fw_identifier)
        else:
            self.pre_install_nic_fw_version = self.flint_dict.get("FW Version", "N/A")

    def _set_running_dual_fw(self):
        """
        Set running_dual_fw field using data from flint_dict.
        """
        if "FW Version(Running)" in self.flint_dict:
            self.running_dual_fw = True
        else:
            self.running_dual_fw = False

    def set_device_properties(self, pci_mst_map: Dict[str, str], lspci_output: str):
        """
        Set detailed properties for the device.

        Method will set rshim interface, mode, PSID,
        firmware versions, and display level.
        """
        # Find RShim management interface
        rshim_pci_regex = rf"({self.bf_pci_address[:-2]}.[0-9]) .+\[15b3:c2d5\]"
        rshim_match = re.search(rshim_pci_regex, lspci_output)
        if rshim_match:
            self.rshim_pci_address = rshim_match.group(1)
        else:
            print(f"WARN: No management interface found for BF device {self.bf_pci_address}")

        self._populate_flint_dict()
        self._set_psid()
        self._set_fw_version()
        self._set_running_dual_fw()
        self._set_bf_mode()
        self._set_nic_pn_sn()
        self._set_privileged_mode()
        self.set_display_level()
        mst_key = self.bf_pci_address.replace("0000:", "")
        self.mst_address = pci_mst_map.get(mst_key)

    def _update_post_installation_fw_versions(self):
        """
        Update the device with post-installation firmware versions.

        Re populate the flint dictionary and update the post-installation version attributes.
        """
        self._populate_flint_dict()
        self.post_install_nic_fw_version = self.flint_dict.get("FW Version", "FW Version(Running)")
        self.post_install_atf_fw_version, self.post_install_uefi_fw_version = get_atf_uefi_versions(self)

        if not self.post_install_bmc_fw_version:
            self.post_install_bmc_fw_version = "N/A"

    def bind_device(self) -> None:
        """
        Bind the BF device to mlx5_core driver if it's not already bound.
        This handles both single and dual-port devices.
        """
        # Extract the base address (everything before the last dot)
        # e.g., "0000:03:00.0" -> base="0000:03:00", port="0"
        base_addr, port = self.bf_pci_address.rsplit('.', 1)

        # We always process port 0 (which is our stored address)
        # and check if port 1 exists

        for port_num in [0, 1]:
            port_addr = f"{base_addr}.{port_num}"

            # Check if this port exists in sysfs
            if not Path(f"/sys/bus/pci/devices/{port_addr}").exists():
                if port_num == 1:
                    # Port 1 doesn't exist, this is a single-port device
                    continue

            # Check if already bound to mlx5_core
            if Path(f"/sys/bus/pci/drivers/mlx5_core/{port_addr}").exists():
                continue

            # Check if bound to another driver
            if Path(f"/sys/bus/pci/devices/{port_addr}/driver").exists():
                continue

            # Bind the port
            bind_cmd = f"echo {port_addr} > /sys/bus/pci/drivers/mlx5_core/bind"
            output, rc = run_command(bind_cmd)

            if rc != 0:
                print(f"WARN: Failed to bind device {port_addr}: {output}")


class MiscLogger:
    """
    A logger class to process and clean logs from the rshim misc buffer.
    """

    def __init__(self):
        self.seen_messages = set()
        self.last_uptime = None
        self.log_buffer = []
        self.header = None

    def split_header_and_logs(self, misc_log: str) -> Tuple[str, str]:
        """
        Split the misc log into header and log sections using the divider.
        """
        divider = "---------------------------------------\n             Log Messages\n---------------------------------------"
        if divider in misc_log:
            header, log_section = misc_log.split(divider, 1)
            return header.strip(), log_section.strip()
        return misc_log.strip(), ""

    def process_log(self, misc_log: str) -> List[str]:
        """
        Process a single misc log entry and return deduplicated log lines with UP_TIME values.
        """
        header, log_section = self.split_header_and_logs(misc_log)
        if self.header is None:
            self.header = header

        uptime_match = re.search(r"UP_TIME\s+(\d+)\(s\)", header)
        current_uptime = int(uptime_match.group(1)) if uptime_match else None

        # Handle UP_TIME reset (e.g., device restart)
        if self.last_uptime is None or (
            current_uptime is not None and current_uptime < self.last_uptime
        ):
            self.seen_messages.clear()

        # Deduplicate new messages from the log section
        new_log_lines = []
        for line in log_section.splitlines():
            if line not in self.seen_messages:
                new_log_lines.append(line)
                self.seen_messages.add(line)

        # Add UP_TIME before new messages
        if (
            current_uptime is not None
            and current_uptime != self.last_uptime
            and new_log_lines
        ):
            new_log_lines.insert(
                0,
                f"----------------------\nUP_TIME         {current_uptime}(s)\n----------------------",
            )

        self.last_uptime = current_uptime
        return new_log_lines

    def get_cleaned_logs(self) -> str:
        """
        Get the cleaned logs with the full header and deduplicated messages.
        """
        cleaned_logs = [self.header] if self.header else []
        cleaned_logs.append("---------------------------------------")
        cleaned_logs.append("             Log Messages")
        cleaned_logs.append("---------------------------------------")
        cleaned_logs.extend(self.log_buffer)
        return "\n".join(cleaned_logs)


class CustomLogger:
    def __init__(self, log_file: str):
        self.terminal = sys.stdout
        self.log_file = open(log_file, "a")

    def write(self, message: str):
        self.terminal.write(message)
        self.log_file.write(message)
        self.log_file.flush()

    def flush(self):
        self.terminal.flush()
        self.log_file.flush()


def setup_logging(filename: str) -> bool:
    """
    Configure logging to save messages to a log file.
    Creates a timestamped directory for this run under the base log directory.
    """
    global BF_LOG_DIR
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    BF_LOG_DIR = os.path.join(BASE_LOG_DIR, f"doca_installer_{timestamp}")

    log_file = os.path.join(BF_LOG_DIR, filename)
    try:
        # Create the base directory if it doesn't exist
        os.makedirs(BASE_LOG_DIR, mode=0o755, exist_ok=True)
        # Create the timestamped directory for this run
        os.makedirs(BF_LOG_DIR, mode=0o755, exist_ok=True)
        sys.stdout = CustomLogger(log_file)
        return True
    except Exception as e:
        print(f"Failed to set up logging: {str(e)}")
        return False


def write_to_file(filename: str, cmd: str, content: str) -> None:
    """
    Write content to file.
    """
    timestamped_filename = f"{filename}_{time.strftime('%Y%m%d_%H%M%S')}.log"
    output_file = os.path.join(BF_LOG_DIR, timestamped_filename)
    try:
        with open(output_file, "w") as f:
            f.write(f"Command: {cmd}\n\n")
            f.write(content)
    except Exception as e:
        print(
            f"ERROR: Failed to write the output of the '{cmd}' command to '{filename}': {e}"
        )


def run_command(
    command: str, timeout: int = DEFAULT_TIMEOUT
) -> Tuple[str, int]:
    """
    Run a shell command and return its output.
    """
    try:
        result = subprocess.run(
            command,
            shell=True,
            timeout=timeout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
        )
        try:
            output = result.stdout.decode("utf-8")
        except UnicodeDecodeError:
            output = result.stdout.decode("utf-8", errors="replace")
        return output, result.returncode

    except subprocess.TimeoutExpired as e:
        partial_output = e.stdout.decode("utf-8", errors="replace") if e.stdout else ""
        timeout_message = f"\nWARN: Command '{command}' timed out after {timeout} seconds"
        combined_output = partial_output + timeout_message
        return combined_output, 1

    except Exception as e:
        print(f"ERROR: Failed to run command '{command}': {str(e)}")
        return str(e), 1


def mstdump_device(device: BFDevice):
    mstdump_device_cmd = f"mstdump --full {device.bf_pci_address}"
    mstdump_device_cmd_output, rc = run_command(mstdump_device_cmd)
    if rc != 0:
        print(f"Failed to run {mstdump_device_cmd}")
    write_to_file(
        f"BF_{device.bf_pci_address.replace(':', '_')}_mstdump_device_cmd_output",
        mstdump_device_cmd,
        mstdump_device_cmd_output,
    )


def parse_bf_devices(lspci_output: str) -> List[BFDevice]:
    """
    Parse BlueField devices from hosts lspci output and create BFDevice objects.

    Args:
        lspci_output (str): Output of the lspci command.

    Returns:
        List[BFDevice]: A list of BFDevice objects with basic
                        initialization - pci address and bf type[BF2/BF3].
    """
    bf_nic_regex = r"([\da-fA-F]{4}:[\da-fA-F]{2}:[\da-fA-F]{2}\.[0-9]) .+\[15b3:(a2dc|a2d6)\]"
    bf_dpu_matches = re.finditer(bf_nic_regex, lspci_output)

    bf_devices = []
    for match in bf_dpu_matches:
        bf_pci_address = match.group(1)
        bf_type = "bf2" if match.group(2) == "a2d6" else "bf3"

        # Skip port 1 for dual-port devices
        if bf_pci_address.endswith(".1"):
            continue
        bf_device = BFDevice(bf_pci_address, bf_type)
        bf_devices.append(bf_device)

    return bf_devices


def get_bf_devices_in_parallel() -> List[BFDevice]:
    """
    Identify all BlueField (BF) devices on the host.
    Fetch will be ran in parallel for each BlueField device, and set each device object fields.
    Method will allow up to max_retries retries for each device.

    Returns:
        List[BFDevice]: A list of BlueField devices with basic initialization.
    """
    bf_devices = []
    start_mft_tool()
    lspci_output, rc = run_command("lspci -D -nn")
    write_to_file("lspci_cmd_output", "lspci -D -nn", lspci_output)

    if rc != 0:
        print(f"ERROR: Failed to run lspci command: {lspci_output}")
        return bf_devices

    # Devices pci-mst map
    pci_mst_map = get_pci_mst_map()

    bf_devices = parse_bf_devices(lspci_output)

    retries = {}
    max_retries = 1

    with ThreadPoolExecutor(max_workers=len(bf_devices)) as executor:
        futures_to_device = {executor.submit(device.set_device_properties, pci_mst_map, lspci_output): device for device in bf_devices}

        while futures_to_device:
            futures_retry = {}
            for future in as_completed(futures_to_device):
                device = futures_to_device[future]
                try:
                    future.result()  # Wait for the task result (raises exception if the task failed)
                except Exception as e:
                    # Retry logic
                    if retries.get(device.bf_pci_address, 0) < max_retries:
                        retries[device.bf_pci_address] = retries.get(device.bf_pci_address, 0) + 1
                        futures_retry[executor.submit(device.set_device_properties, pci_mst_map, lspci_output)] = device
                    else:
                        print(f"ERROR: Failed fetching details of {device.bf_pci_address} DPU after {max_retries} retries")

            # Update futures_to_device with new retries
            futures_to_device = futures_retry

    return bf_devices


def dump_mlxconfig(device: BFDevice) -> Tuple[Optional[str], Optional[str]]:
    """
    Run full mlxconfig query on BF DPU device
    """
    mlxconfig_cmd = f"mlxconfig -d {device.bf_pci_address} -e q"
    mlxconfig_output, rc = run_command(mlxconfig_cmd, MLX_CONFIG_FULL_TIMEOUT)
    write_to_file(
        f"BF_{device.bf_pci_address.replace(':', '_')}_mlxconfig_cmd_output",
        mlxconfig_cmd,
        mlxconfig_output,
    )

    return mlxconfig_output, rc


def get_rshim_devices() -> Dict[str, str]:
    """
    Get the corresponding RSHIM device for each BF DPU.
    """
    rshim_pci_mapping = {}
    rshim_pci_regx = (
        r"DEV_NAME\s+pcie-([\da-fA-F]{4}:[\da-fA-F]{2}:[\da-fA-F]{2}\.[0-9])"
    )

    # Find all rshim file devices
    rshim_devices = [
        file.name
        for file in os.scandir("/dev")
        if file.name.startswith("rshim")
    ]

    for rshim_device in rshim_devices:
        source_file = f"/dev/{rshim_device}/misc"
        dest_file = f"{BF_LOG_DIR}/BF_{rshim_device}_misc_{time.strftime('%Y%m%d_%H%M%S')}"
        try:
            shutil.copy2(source_file, dest_file)
            # print(f"INFO: Copied {source_file} to {dest_file}")
            with open(source_file, "r") as misc_file:
                misc_content = misc_file.read()
            rshim_pci_match = re.search(rshim_pci_regx, misc_content)
            if rshim_pci_match:
                rshim_pci_mapping[rshim_pci_match.group(1)] = rshim_device
            os.remove(dest_file)
        except Exception as e:
            print(
                f"ERROR: Error reading from /dev/{rshim_device}/misc: {str(e)}"
            )

    return rshim_pci_mapping


def set_rshim_force_mode() -> None:
    """
    Method verify if rshim FORCE MODE is not set, and set it if not.
    """
    pattern = r"^\s*(#)?\s*FORCE_MODE\s+([01])"
    found = False
    mode = None
    is_commented = None
    with open("/etc/rshim.conf", "r") as file:
        for line in file:
            match = re.match(pattern, line)
            if match:
                found = True
                # Extract groups
                is_commented = bool(match.group(1))  # True if '#' exists
                mode = int(match.group(2))  # Extract '0' or '1'
    command = None
    replace_with_commented_cmd = r"sed -i 's/^#\(FORCE_MODE[[:space:]]*\)0/\1 1/' /etc/rshim.conf"
    replace_no_commented_cmd = r"sed -i 's/^\(FORCE_MODE[[:space:]]*\)0/\11/' /etc/rshim.conf"
    uncommand_cmd = r"sed -i '/FORCE_MODE[[:space:]]*1/s/^#\s*//' /etc/rshim.conf"
    if found:
        if mode == 0:
            if is_commented:
                command = replace_with_commented_cmd
            else:
                command = replace_no_commented_cmd
        elif is_commented:
            command = uncommand_cmd
    else:
        command = r"echo 'FORCE_MODE 1' >> /etc/rshim.conf"

    if command:
        _, _ = run_command(command)


def restart_rshim_service() -> None:
    """
    Restart the RSHIM service on a host.
    """
    set_rshim_force_mode()

    for i in range(MAX_RETRIES):
        restart_rshim_cmd = "systemctl restart rshim.service"
        restart_rshim_cmd_output, rc = run_command(
            restart_rshim_cmd, RSHIM_SERVICE_RESTART_TIMEOUT
        )
        if rc != 0:
            print(
                f"ERROR: Failed to restart RSHIM service: {restart_rshim_cmd_output}"
            )
            time.sleep(60)
            continue

        show_rshim_service_cmd = "systemctl status rshim.service"
        show_rshim_service_cmd_output, rc = run_command(show_rshim_service_cmd)
        if (
            "failed" in show_rshim_service_cmd_output.lower()
            or "error" in show_rshim_service_cmd_output.lower()
        ):
            print(
                f"ERROR: Found error in RSHIM service restart: {show_rshim_service_cmd_output}"
            )
            time.sleep(60)
        else:
            time.sleep(60)  # Allow service to fully initialize
            break


def get_atf_uefi_versions(
    device: BFDevice,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Get the ATF and UEFI firmware versions for a given BF DPU
    by reading from rshim console for 30 seconds and saving output to a file.
    """
    if device.bf_type == "bf3":
        # Try extract ATF and UEFI versions using fw register
        atf_version = get_version_from_register(device, "ATF")
        uefi_version = get_version_from_register(device, "UEFI")
        if atf_version and uefi_version:
            return atf_version, uefi_version

    console_path = f"/dev/{device.rshim_device}/console"
    output_file = f"{BF_LOG_DIR}/BF_{device.bf_pci_address.replace(':', '_')}_{device.rshim_device}_console_{time.strftime('%Y%m%d_%H%M%S')}.log"

    atf_ver_regex = r"BL2:.+\(release\):([\w.-]+)"
    uefi_ver_regex = r"UEFI firmware\s+\(version BlueField:([\w.-]+)\s"
    try:
        # Capture rshim console output
        with open(console_path, "rb", buffering=0) as console, open(
            output_file, "wb"
        ) as log_file:
            os.set_blocking(console.fileno(), False)

            # Reset BF DPU
            sw_reset_bf_device(device)

            start_time = time.time()
            while time.time() - start_time < CONSOLE_READ_TIMEOUT:
                try:
                    data = os.read(console.fileno(), 1024)
                    if data:
                        log_file.write(data)
                        log_file.flush()
                except BlockingIOError:
                    time.sleep(0.01)

        # Save captured output
        with open(output_file, "r", errors="ignore") as log_file:
            console_content = log_file.read()

        # Extract ATF and UEFI FW versions
        atf_match = re.search(atf_ver_regex, console_content)
        uefi_match = re.search(uefi_ver_regex, console_content)

        atf_version = atf_match.group(1) if atf_match else None
        uefi_version = uefi_match.group(1) if uefi_match else None

        return atf_version, uefi_version

    except Exception as e:
        print(f"ERROR: Error processing console output: {str(e)}")
        return None, None


def sw_reset_bf_device(device: BFDevice) -> None:
    """
    Reset ATF/UEFI firmware on a BF DPU.
    """
    sw_reset_bf_device_cmd = (
        f"echo 'SW_RESET 1' > /dev/{device.rshim_device}/misc"
    )
    sw_reset_bf_device_cmd_output, rc = run_command(sw_reset_bf_device_cmd)

    if rc != 0:
        print(
            f"ERROR: Failed to reset ATF/UEFI firmware on BF device {device.bf_pci_address}: {sw_reset_bf_device_cmd_output}"
        )


def start_mft_tool() -> bool:
    """
    Start MFT firmware tool.
    """
    mst_cmd = "mst start"
    mst_cmd_output, rc = run_command(mst_cmd)

    return rc == 0


def display_bf_device_basic_info(bf_devices: List[BFDevice]) -> None:
    """
    Display BF basic HW info, e.g., P/N, S/N, PCI address.
    """
    header = [
        "PSID",
        "Part Num",
        "Serial Num",
        "PCI Addr",
        "Mode",
        "Privileged",
    ]
    table = []

    for bf_device in bf_devices:
        table.append(
            [
                bf_device.psid,
                bf_device.part_number,
                bf_device.serial_number,
                bf_device.bf_pci_address,
                bf_device.mode,
                bf_device.nic_fw_privileged_state,
            ]
        )

    if bf_devices:
        print(f"\n{len(bf_devices)} BF DPUs:")
        print(tabulate(table, headers=header, tablefmt="fancy_grid"))


def display_bf_device_info_with_rshim(bf_devices: List[BFDevice]) -> None:
    """
    Display BF basic HW info, e.g., P/N, S/N, PCI address, RSHIM device.
    """
    header = [
        "PSID",
        "Part Num",
        "Serial Num",
        "PCI Addr",
        "RSHIM Dev",
        "Mode",
    ]
    table = []

    for bf_device in bf_devices:
        table.append(
            [
                bf_device.psid,
                bf_device.part_number,
                bf_device.serial_number,
                bf_device.bf_pci_address,
                bf_device.rshim_device,
                bf_device.mode,
            ]
        )

    if bf_devices:
        print(f"\n{len(bf_devices)} BF DPUs:")
        print(tabulate(table, headers=header, tablefmt="fancy_grid"))
        print("\n")


def display_bf_device_full_info(
    bf_devices: List[BFDevice], stage="Pre Upgrade"
) -> None:
    """
    Display BF devices full information.
    """
    header = [
        "PSID",
        "Part Num",
        "Serial Num",
        "PCI Addr",
        "RSHIM Dev",
        "Mode",
        "Privileged",
        f"{stage} ATF",
        f"{stage} UEFI",
        f"{stage} NIC-FW",
        f"{stage} BMC",
        f"{stage} CEC",
    ]
    table = []

    for bf_device in bf_devices:
        row = [
            bf_device.psid,
            bf_device.part_number,
            bf_device.serial_number,
            bf_device.bf_pci_address,
            bf_device.rshim_device,
            bf_device.mode,
            bf_device.nic_fw_privileged_state,
            bf_device.pre_install_atf_fw_version,
            bf_device.pre_install_uefi_fw_version,
            bf_device.pre_install_nic_fw_version,
            bf_device.pre_install_bmc_fw_version,
            bf_device.pre_install_cec_fw_version,
        ]
        table.append(row)

    if bf_devices:
        print(f"\n{len(bf_devices)} BF DPUs:")
        print(tabulate(table, headers=header, tablefmt="fancy_grid",missingval="N/A"))


def check_bf_devices_prerequisites(
    bf_devices: List[BFDevice],
    specific_psid: str = None,
    report_mode="ERROR",
) -> Tuple[bool, List[BFDevice]]:
    """
    Method will check all devcies for the defined pre requisites.
        1. If specific PSID is defined, it will check only that PSID
        2. Device must be in privilaged mode
        3. RSHIM pci address must be assosieted with a BF device
    Any device that stand for those, will be returned to the caller

    Args:
        bf_devices (List[BFDevice]): List of avaliable devices to check
        bundle_type (str): bf-bundle or bf-fwbundle
        specific_psid (str, optional): PSID of the device to check. Defaults to None.

    Returns:
        Tuple[bool, List[BFDevice]]:
            True and list of BF devices to upgrade
            False and empty list if there are no BF devices that meets the prerequisites
    """

    target_bf_devices = bf_devices
    if specific_psid:
        target_bf_devices = [
            bf3_device
            for bf3_device in target_bf_devices
            if bf3_device.psid == specific_psid
        ]
        if len(target_bf_devices) == 0:
            print(f"ERROR: Failed to find BF device with PSID: {specific_psid}")
            return False, []

    total_target_bf_devices = len(target_bf_devices)

    bf_devices_in_privileged_mode = [
        bf_device
        for bf_device in target_bf_devices
        if bf_device.nic_fw_privileged_state
    ]
    bf_devices_in_restricted_mode = [
        bf_device
        for bf_device in target_bf_devices
        if bf_device not in bf_devices_in_privileged_mode
    ]
    if len(bf_devices_in_privileged_mode) < total_target_bf_devices:
        print(
            f"INFO: {total_target_bf_devices - len(bf_devices_in_privileged_mode)} out of {total_target_bf_devices} BF DPUs are NOT in PRIVILEGED mode!"
        )
        for bf_device in bf_devices_in_restricted_mode:
            print(
                f"INFO: BF Device {bf_device.bf_pci_address} not in PRIVILEGED mode - and will not upgrade!"
            )
        target_bf_devices = bf_devices_in_privileged_mode
    total_target_bf_devices = len(target_bf_devices)

    bf_devices_with_rshim_port = [
        bf_device
        for bf_device in target_bf_devices
        if bf_device.rshim_pci_address
    ]
    if len(bf_devices_with_rshim_port) < total_target_bf_devices:
        print(
            f"INFO: Host Access is not enabled for {total_target_bf_devices - len(bf_devices_with_rshim_port)} out of {total_target_bf_devices} BF DPUs."
        )
        for bf_device in [
            bf_device
            for bf_device in target_bf_devices
            if bf_device not in bf_devices_with_rshim_port
        ]:
            print(
                f"INFO: BF Device {bf_device.bf_pci_address} does not have an attached rshim pci - and will not upgrade!"
            )
        target_bf_devices = bf_devices_with_rshim_port

    if target_bf_devices:
        return True, target_bf_devices
    else:
        return False, []


def check_bf_rshim_devices(
    bf_devices: List[BFDevice], specific_rshim_device: str = None
) -> Tuple[bool, List[BFDevice]]:
    """
    Check if all BF DPUs have RSHIM ports and console ports are working.
    """
    total_bf_devices = len(bf_devices)
    devices_to_verify = []
    for i in range(MAX_RETRIES):
        restart_rshim_service()
        rshim_pci_mapping = get_rshim_devices()
        for bf_device in bf_devices:
            bf_device.rshim_device = rshim_pci_mapping.get(
                bf_device.rshim_pci_address
            )
            if not bf_device.rshim_device:
                print(
                    f"ERROR: Failed to find RSHIM port for BF device {bf_device.bf_pci_address}!"
                )
            else:
                bf_device.set_display_level()

        devices_to_verify = [
            bf_device for bf_device in bf_devices if bf_device.rshim_device
        ]

        if specific_rshim_device:
            devices_to_verify = [
                bf_device
                for bf_device in devices_to_verify
                if bf_device.rshim_device == specific_rshim_device
            ]
            if not devices_to_verify:
                print(
                    f"ERROR: Got request to upgrade specific RSHIM device {specific_rshim_device}!\n"
                )
                print(
                    f"ERROR: Failed to find device with RSHIM port {specific_rshim_device}!\n"
                )
                return False, []
            total_bf_devices = len(devices_to_verify)

        if len(devices_to_verify) < total_bf_devices:
            print(
                f"ERROR: {total_bf_devices - len(devices_to_verify)} out of {total_bf_devices} BF DPUs missing RSHIM port! Retrying..."
            )
            continue

        for bf_device in devices_to_verify:
            bf_device.pre_install_atf_fw_version, bf_device.pre_install_uefi_fw_version = (
                get_atf_uefi_versions(bf_device)
            )
            if not bf_device.pre_install_atf_fw_version:
                print(
                    f"ERROR: Failed to collect ATF/UEFI FW versions for BF device {bf_device.bf_pci_address}."
                )

            # Try to get BMC/CEC versions
            bf_device.pre_install_bmc_fw_version = get_version_from_register(bf_device, "BMC")
            bf_device.pre_install_cec_fw_version = get_version_from_register(bf_device, "CEC")

        bf_devices_with_atf_fw_ver = [
            bf_device
            for bf_device in devices_to_verify
            if bf_device.pre_install_atf_fw_version
        ]
        if len(bf_devices_with_atf_fw_ver) < total_bf_devices:
            print(
                f"ERROR: Failed to collect ATF/UEFI FW versions for {total_bf_devices - len(bf_devices_with_atf_fw_ver)} out of {total_bf_devices} BF devices! Retrying..."
            )
            continue
        else:
            return True, bf_devices_with_atf_fw_ver

    return False, []


def verify_device_fw_state(bf_devices: List[BFDevice]) -> List[BFDevice]:
    """
    Check if all BF DPUs are ready for FW upgrade - Not running burned FW that were not loaded.
    """
    device_list_ready_for_nic_fw_burn = [
        bf_device for bf_device in bf_devices if not bf_device.running_dual_fw
    ]
    device_to_reset_ir = [
        bf_device
        for bf_device in bf_devices
        if bf_device not in device_list_ready_for_nic_fw_burn
    ]
    for device in device_to_reset_ir:
        print(
            f"WARN: Device {device.bf_pci_address} is running unloaded burned FW, running flint image reactivate"
        )
        ir_cmd = f"flint -d {device.bf_pci_address} ir"
        ir_cmd_output, rc = run_command(ir_cmd)
        if rc == 0:
            device_list_ready_for_nic_fw_burn.append(device)
        else:
            print(
                f"ERROR: Failed to reset IR on device {device.bf_pci_address}: {ir_cmd_output}"
            )

    return device_list_ready_for_nic_fw_burn


def check_mst_status() -> bool:
    """
    Check the status of the MST (Multi-Function Switch) and verify if all cards are available.

    Returns:
        bool: True if all cards are available, False otherwise.
    """
    cards_lines = []
    require_sleep = False
    mst_status_command = "mst status -v"
    mst_status_command_output, rc = run_command(mst_status_command)
    if rc != 0:
        print("WARN: Failed to fetch mst status -v")
        print("MST Content:")
        print(mst_status_command_output)
        return True, True
    lines = [
        line for line in mst_status_command_output.splitlines() if line != ""
    ]
    for i, line in enumerate(lines):
        if line.startswith("PCI devices:") and i < len(lines) - 3:
            cards_lines = lines[i + 3:]
            break
    for line in cards_lines:
        if len(line.split()) != 6 or any(
            item == "NA" for item in line.split()
        ):
            require_sleep = True
            break
    if require_sleep:
        print("WARN: One of the cards is not avaliable in MST status")
        print("MST Content:")
        print(mst_status_command_output)
    return require_sleep


def verify_mst_status() -> None:
    """
    Verify MST status and cards.
    """
    mst_start_command = "mst start"
    mst_status_command_output, rc = run_command(mst_start_command)
    if rc != 0:
        print("WARN: Failed to start MST \n")
        return
    time.sleep(5)
    for iteration in range(3):
        require_sleep = check_mst_status()
        if iteration == 2:
            break
        elif require_sleep:
            time.sleep(30)
        else:
            break


def get_pci_mst_map() -> Dict[str, str]:
    """
    Get the PCI MST map.

    This function fetches the status of all PCI devices.
    It then creates a dictionary mapping each PCI device to its corresponding MST address.

    Returns:
        Dict[str, str]: A dictionary where the keys are PCI device addresses and the values are MST addresses.
    """
    mst_start_command = "mst start"
    mst_status_command_output, rc = run_command(mst_start_command)
    if rc != 0:
        print("WARN: Failed to start MST \n")
        return
    pci_mst_map = {}
    mst_status_command = "mst status -v"
    mst_status_command_output, rc = run_command(mst_status_command)
    write_to_file(
        "mst_status_-v_for_map", mst_status_command, mst_status_command_output
    )
    if rc != 0:
        print("WARN: Failed to fetch mst status -v")
        print("MST Content:")
        print(mst_status_command_output)
        return pci_mst_map

    lines = mst_status_command_output.splitlines()
    pci_to_mst_map = {}
    for line in lines:
        columns = line.split()
        if len(columns) >= 3:
            device_type, mst_value, pci_value = columns[:3]
            if mst_value.startswith("/dev/mst/") and ":" in pci_value:
                pci_to_mst_map[pci_value] = mst_value
    return pci_to_mst_map


def get_version_from_register(
    device: BFDevice,
    required_component: str
) -> Optional[str]:
    """
    Generic function to retrieve version from MISOC register using mlxreg command.

    :param device: The BFDevice object containing device information
    :param required_component: Type of version to query ("ATF", "UEFI", "BMC", or "CEC")
    :return: Version string if successful, None if unavailable
    """
    # Define version metadata for MISOC register fetching
    COMPONENT_LOOKUP = {
        "ATF": {"type": 0, "num_fields": 17},
        "UEFI": {"type": 1, "num_fields": 17},
        "BMC": {"type": 2, "num_fields": 24},
        "CEC": {"type": 3, "num_fields": 20},
    }

    # Only supported for BF3 devices
    if device.bf_type != "bf3":
        return None

    if required_component not in COMPONENT_LOOKUP:
        print(f"ERROR: Unknown version type: {required_component}")
        return None

    config = COMPONENT_LOOKUP[required_component]

    try:
        device_id = f"BF_{device.bf_pci_address.replace(':', '_')}"

        # Run mlxreg command for the specific firmware type
        cmd = f"mlxreg -d {device.mst_address} --reg_name MISOC --get --index type={config['type']}"
        output = subprocess.check_output(
            cmd.split(), stderr=subprocess.PIPE, universal_newlines=True
        )

        # Log the command output
        write_to_file(
            f"{device_id}_misoc_{required_component.lower()}",
            cmd,
            output
        )

    except subprocess.CalledProcessError as e:
        # Log error output if mlxreg fails
        write_to_file(
            f"{device_id}_mlxreg_error_{required_component.lower()}",
            e.cmd,
            e.stderr
        )
        return None

    # Extract hex values from version fields
    hex_string = ""
    for i in range(config["num_fields"]):
        pattern = rf"version\[{i}\]"
        matches = re.findall(rf"{pattern}.*", output, flags=re.IGNORECASE)

        if matches:
            tmp = matches[0].split("|")[1]
            hex_string += tmp[9:19]

    # Convert hex string to readable string
    try:
        version_str = bytes.fromhex(hex_string).decode("utf-8", errors="ignore").strip('\x00')
        return version_str
    except ValueError as e:
        # Log decoding error
        write_to_file(
            f"{device_id}_hex_decode_error_{required_component.lower()}",
            "Hex decoding",
            str(e)
        )
        return None


def extract_and_set_bmc_version(device: BFDevice, text: str):
    """
    Extract BMC firmware version from text and set it to the device if not already set.

    Args:
        device (BFDevice): The device to update
        text (str): The text to search for BMC firmware version
    """
    bmc_match = re.search(r"BMC firmware was updated to: ([0-9]+\.[0-9]+-[0-9]+)", text)
    if bmc_match and not device.post_install_bmc_fw_version:
        device.post_install_bmc_fw_version = bmc_match.group(1).strip()


def verify_device(
    bf_device: BFDevice,
    progress: dict = None,
    lock: threading.Lock = None,
    TIMEOUT=TIME_FOR_DEVICE_TO_BOOT,
) -> None:
    """
    Verify if a device is ready post install.
    Method will track rshim misc buffer if device is up and running
    Method will also track if power cycle is required

    Args:
        bf_device (BFDevice): The device to verify.
        progress (dict, optional): A dictionary to store the progress of the verification. Defaults to None.
        lock (threading.Lock, optional): A lock to synchronize access to the progress dictionary. Defaults to None.
    """
    logger = MiscLogger()
    file_output_name = f"BF_{bf_device.bf_pci_address.replace(':', '_')}_cat_rshim_misc_cmd"

    expected_status = "DPU is ready" if bf_device.mode == "DPU" else "In Enhanced NIC mode"
    power_cycle_string = "power cycle is required"
    
    poll_interval = 0.25  # seconds between checks
    max_iterations = int(TIMEOUT / poll_interval)

    for _ in range(max_iterations):
        # Check if we've been asked to terminate early (global timeout reached)
        if progress and lock:
            with lock:
                if progress[bf_device]["monitor_done"]:
                    return

        misc_command = f"cat /dev/{bf_device.rshim_device}/misc"
        misc_command_output, rc = run_command(misc_command)

        # Process the misc log using MiscLogger
        new_logs = logger.process_log(misc_command_output)
        if new_logs:
            logger.log_buffer.extend(new_logs)

        if power_cycle_string in misc_command_output:
            bf_device.required_power_cycle = True

        if expected_status in misc_command_output and rc == 0:
            cleaned_logs = logger.get_cleaned_logs()
            write_to_file(file_output_name, misc_command, cleaned_logs)

            # Update progress safely if lock is provided
            if lock:
                with lock:
                    progress[bf_device]["monitor_done"] = True
                    progress[bf_device]["reason"] = "Success"
                    # Check for BMC firmware update in the cleaned logs
                    extract_and_set_bmc_version(bf_device, cleaned_logs)
            bf_device.dpu_ready = True
            return

        time.sleep(poll_interval)

    cleaned_logs = logger.get_cleaned_logs()
    write_to_file(file_output_name, misc_command, cleaned_logs)

    # Finalize progress update after timeout if running as part of parallel flow
    if lock:
        with lock:
            progress[bf_device]["monitor_done"] = True
            progress[bf_device]["reason"] = "Device verification timed out - DPU did not become ready"
            extract_and_set_bmc_version(bf_device, cleaned_logs)
    return


def kill_bfb_install(rshim: str) -> None:
    """
    Terminate bfb-install run of a given rshim.

    Args:
        rshim (str): rshim number (i.e rshim0)
    """
    main_script_name = os.path.basename(sys.argv[0])
    current_pid = os.getpid()
    for proc in psutil.process_iter(["pid", "name", "cmdline"]):
        if (
            "bfb-install" in proc.info["name"]
            and rshim in proc.info["cmdline"]
        ):
            if (
                proc.info["pid"] == current_pid
                or main_script_name in proc.info["cmdline"]
            ):
                continue
            run_command(f'kill -9 {proc.info["pid"]}')

def install_bfb(
    bf_device: BFDevice, bf_path: str, progress: dict, lock: threading.Lock
) -> bool:
    """
    Installs BFB on a given BFDevice.

    Args:
        bf_device (BFDevice): The BFDevice to install BFB on.
        bf_path (str): The path to the BFB.
        progress (dict): A dictionary to track the progress of the installation.
        lock (threading.Lock): A lock to ensure thread safety in order to update the progress.

    Returns:
        bool: True if the installation was successful, False otherwise.
    """
    bfb_install_output = ""
    rc = 1
    failure_reason = ""
    
    try:
        # Check if early termination was requested
        with lock:
            if progress[bf_device]["install_done"]:
                return False
                
        bf_device.set_display_level()
        bfb_install_cmd = (
            f"{BFB_INSTALL_PATH} -v -b {bf_path} -r {bf_device.rshim_device}"
        )
        if bf_device.device_config_file:
            bfb_install_cmd += f" -c {bf_device.device_config_file}"

        bfb_install_output, rc = run_command(
            bfb_install_cmd, BUDNLE_INSTALL_TIMEOUT
        )
        write_to_file(
            f"BF_{bf_device.bf_pci_address.replace(':', '_')}_bfb_install_output",
            bfb_install_cmd,
            bfb_install_output,
        )
    except Exception as e:
        failure_reason = f"BFB installation failed: {str(e)}"
        print(
            f"ERROR: Failed to install BFB on {bf_device.bf_pci_address}: {str(e)}"
        )
        rc = 1
    finally:
        # Check again for early termination
        with lock:
            if progress[bf_device]["install_done"]:
                return False
        
        # If we have an installation error/timeout        
        if rc != 0:
            if "timed out" in bfb_install_output:
                failure_reason = "BFB installation command timed out"
                kill_bfb_install(bf_device.rshim_device)

                time.sleep(5)  # Wait 5 seconds after kill
                bf_device.bind_device()

                sw_reset_bf_device(bf_device)
                # Check one more time before calling verify_device
                if not progress[bf_device]["install_done"]:
                    rc = verify_device(bf_device=bf_device, TIMEOUT=120)
            elif not failure_reason:
                failure_reason = f"BFB installation failed with return code {rc}"
                
        # Update progress safely
        with lock:
            progress[bf_device]["install_done"] = True
            if rc != 0 and not progress[bf_device]["reason"]:
                progress[bf_device]["reason"] = failure_reason
            extract_and_set_bmc_version(bf_device, bfb_install_output)

        return rc == 0

def flash_rshim_misc(bf_device: BFDevice) -> None:
    """
    Flash rshim misc file.
    """
    clean_on_read_true = f"echo 'CLEAR_ON_READ 1' > /dev/{bf_device.rshim_device}/misc"
    clean_on_read_false = f"echo 'CLEAR_ON_READ 0' > /dev/{bf_device.rshim_device}/misc"
    cat_rshim_misc = f"cat /dev/{bf_device.rshim_device}/misc"
    _, _ = run_command(clean_on_read_true)
    _, _ = run_command(cat_rshim_misc)
    _, _ = run_command(cat_rshim_misc)
    _, _ = run_command(clean_on_read_false)


def check_bfb_install_log_for_ready_status(bf_device: BFDevice) -> bool:
    """
    Check if the bfb-install log contains the expected ready status message.
    This is used as a fallback when the misc monitoring doesn't catch the status.
    
    Args:
        bf_device (BFDevice): The device to check
        
    Returns:
        bool: True if the expected status message is found in the log
    """
    expected_status = "DPU is ready" if bf_device.mode == "DPU" else "In Enhanced NIC mode"
    log_file_path = os.path.join(BF_LOG_DIR, f"BF_{bf_device.bf_pci_address.replace(':', '_')}_bfb_install_output.log")
    
    try:
        if os.path.exists(log_file_path):
            with open(log_file_path, 'r') as f:
                log_content = f.read()
                if expected_status in log_content:
                    return True
    except Exception as e:
        # If we can't read the log file, assume the status wasn't found
        pass
        
    return False


def install_bundle_orchestration(
    bf_devices: List[BFDevice], image_path: str
) -> Tuple[bool, List[BFDevice]]:
    """
    Install bundle orchestration on a list of BF devices.
    First, method will flash misc content.
    Method will then install and will verify if device is ready
    post install in parallel for all available device in list.

    Args:
        bf_devices (List[BFDevice]): A list of BF devices.
        image_path (str): The path to the image.

    Returns:
        Tuple[bool, List[BFDevice]]:
            - Status of all devices (True if all are ready, False otherwise).
            - A list of devices that are ready.
    """
    flash_threads = []
    for device in bf_devices:
        thread = threading.Thread(
            target=flash_rshim_misc,
            args=(device,),
            name=f"FlashThread-{device.rshim_device}"
        )
        flash_threads.append(thread)
        thread.start()

    # Wait for all flash operations to complete
    for thread in flash_threads:
        thread.join()

    progress = {
        device: {"install_done": False, "monitor_done": False, "reason": ""}
        for device in bf_devices
    }
    overall_progress = tqdm(
        total=len(bf_devices),
        desc=f"{len(bf_devices):>4} BF install DPU(s)",
        position=0,
    )

    threads = []
    completed_devices = set()  # Track devices that have been printed as completed
    lock = threading.Lock()  # Lock for thread-safe access to shared data

    for device in bf_devices:
        # Create threads for both tasks
        t1 = threading.Thread(
            target=install_bfb,
            args=(device, image_path, progress, lock),
            name=f"InstallThread-{device.rshim_device}"
        )
        t2 = threading.Thread(
            target=verify_device,
            args=(device, progress, lock),
            name=f"VerifyThread-{device.rshim_device}"
        )
        t3 = threading.Thread(
            target=dump_mlxconfig,
            args=(device,),
            name=f"DumpMlxconfigThread-{device.rshim_device}"
        )
        threads.append((t1, t2, t3))

        # Start both threads
        t1.start()
        t2.start()
        t3.start()

    # Set a maximum overall timeout for the installation process
    max_installation_time = time.time() + TIME_FOR_DEVICE_TO_BOOT + 120  # Use the same timeout as verify_device
    
    # Monitor progress of all units
    while (time.time() < max_installation_time and 
           any(not (p["install_done"] and p["monitor_done"]) for p in progress.values())):
        time.sleep(1)

        for device in bf_devices:
            if (
                progress[device]["install_done"]
                and progress[device]["monitor_done"]
                and device not in completed_devices
            ):
                print(f"{overall_progress.n + 1}/{len(bf_devices)} done: Device {device.bf_pci_address} is ready")
                completed_devices.add(device)
                overall_progress.update(1)
        overall_progress.refresh()

    # Mark any remaining devices as done to ensure the threads can complete
    with lock:
        for device in bf_devices:
            if not progress[device]["install_done"] or not progress[device]["monitor_done"]:
                # Set reason for timeout if not already set
                if not progress[device]["reason"]:
                    progress[device]["reason"] = "Operation timed out"
                progress[device]["install_done"] = True
                progress[device]["monitor_done"] = True
    
    # Wait for all threads to finish (they should finish quickly now)
    for t1, t2, t3 in threads:
        t1.join()
        t2.join()
        t3.join()

    overall_progress.close()

    # Check for devices that completed but weren't marked as ready
    # This handles cases where the installation succeeded but misc monitoring didn't catch the ready status
    # By this point, all devices have monitor_done=True (either naturally or forced by timeout)
    for device in bf_devices:
        if not device.dpu_ready:
            # Check if the bfb-install log contains the ready status
            if check_bfb_install_log_for_ready_status(device):
                device.dpu_ready = True
                progress[device]["reason"] = "Device ready status found in bfb-install log (fallback check)"
                print(f"Device {device.bf_pci_address}: Ready status confirmed via bfb-install log")

    # Identify which devices timed out or didn't complete successfully
    devices_ready = [bf_device for bf_device in bf_devices if bf_device.dpu_ready]
    devices_not_ready = [bf_device for bf_device in bf_devices if not bf_device.dpu_ready]
    
    # Report final status
    if len(devices_ready) < len(bf_devices):
        print(f"\nInstallation is over, {len(devices_ready)}/{len(bf_devices)} devices finished.")
        for device in devices_not_ready:
            reason = progress[device]["reason"] or "Unknown failure"
            print(f"Device {device.bf_pci_address} did not finish successfully: {reason}")
    
    all_devices_ready = len(devices_ready) == len(bf_devices)
    return all_devices_ready, devices_ready


def _extract_component(mlx_mkbfb_path, component, image_path, temp_dir):
    """
    Extract a single component from a BlueField firmware image using mlx-mkbfb.

    This function extracts a specified component (info-v0, bl33-v0, or bl2-v0) from
    the provided firmware image file. The extracted component is saved to a file named
    'dump-{component}' in the specified temporary directory.

    Args:
        mlx_mkbfb_path (str): Full path to the mlx-mkbfb utility
        component (str): Name of the component to extract (e.g., 'info-v0')
        image_path (str): Path to the firmware image file
        temp_dir (str): Directory where extracted components will be saved

    Returns:
        bool: True if extraction was successful and the dump file exists,
              False if extraction failed or the file wasn't created

    Example:
        extract_component('/usr/bin/mlx-mkbfb', 'info-v0', '/path/to/image.bfb', '/tmp/extract_dir')
    """
    cmd = f"{mlx_mkbfb_path} -x -n {component} {image_path}"
    output, rc = run_command(cmd)

    if rc != 0:
        print(f"Warning: Failed to extract {component} - component may not exist in image or extraction not supported.")
        return False
    else:
        dump_file_path = os.path.join(temp_dir, f"dump-{component}")
        if os.path.exists(dump_file_path):
            return True
        else:
            return False


def extract_firmware_versions(image_path) -> bool:
    """
    Extract and parse firmware versions from a BlueField firmware image.

    This function extracts multiple components (info-v0, bl33-v0, bl2-v0) from the
    provided firmware image in parallel using ThreadPoolExecutor. After extraction,
    it processes the files sequentially to parse firmware version information and
    stores the results in the global TARGET_FW_VERSIONS dictionary.

    The function creates a temporary directory for extraction, changes to that directory
    during processing, and cleans up afterwards regardless of success or failure.

    Args:
        image_path (str): Path to the firmware image file

    Returns:
        bool: True if info-v0 component was successfully extracted and processed,
              False if the image doesn't exist, mlx-mkbfb is not available, or
              info-v0 extraction failed

    Notes:
        - Requires the mlx-mkbfb utility to be installed on the system
        - Updates the global TARGET_FW_VERSIONS dictionary with parsed version information
        - Temporary files are automatically cleaned up after processing
    """
    if not os.path.exists(image_path):
        print(f"Error: Image file {image_path} does not exist.")
        return False

    current_dir = os.getcwd()
    temp_dir = tempfile.mkdtemp(prefix="bf_fw_extract_")
    os.chdir(temp_dir)

    try:
        # Check if mlx-mkbfb exists
        cmd = "which mlx-mkbfb"
        mlx_mkbfb_path, rc = run_command(cmd)
        if rc != 0:
            print("Warning: mlx-mkbfb does not exist on host - can't extract versions.")
            return False

        mlx_mkbfb_path = mlx_mkbfb_path.strip()

        # Extract components in parallel
        components = ["info-v0", "bl33-v0", "bl2-v0"]
        component_results = {component: False for component in components}

        with ThreadPoolExecutor(max_workers=3) as executor:
            # Create a list of extraction tasks
            extraction_tasks = {
                executor.submit(_extract_component, mlx_mkbfb_path, component, image_path, temp_dir): component
                for component in components
            }

            # Process results as they complete
            for future in as_completed(extraction_tasks):
                component = extraction_tasks[future]
                try:
                    component_success = future.result()
                    component_results[component] = component_success
                except Exception as e:
                    print(f"Error extracting {component}: {e}")
                    component_results[component] = False

        # Parse firmware info if info-v0 was successfully extracted
        if component_results["info-v0"]:
            _parse_firmware_info(temp_dir)
            return True
        else:
            print("Warning: info-v0 component could not be extracted from the image.")
            # Still try to parse other components if they were extracted
            if any(component_results.values()):
                _parse_firmware_info(temp_dir)
            return False

    except Exception as e:
        print(f"Error in firmware extraction process: {e}")
        return False

    finally:
        # Always return to original directory and clean up
        os.chdir(current_dir)
        shutil.rmtree(temp_dir)


def _parse_firmware_info(data_temp_dir):
    """
    Parse firmware information from extracted component files and populate version data.

    This function reads multiple firmware component files from the specified temporary
    directory and extracts version information to populate the global TARGET_FW_VERSIONS
    dictionary. It processes:

    1. dump-info-v0: JSON file containing component details and versions
    2. dump-bl33-v0: UEFI firmware component (processed with zcat/gunzip and strings)
    3. dump-bl2-v0: ATF firmware component (processed with strings)

    The function updates the global TARGET_FW_VERSIONS dictionary with version
    information for all firmware components found in the extracted files.

    Args:
        data_temp_dir (str): Directory path containing the extracted firmware component files

    Error Handling:
        - Prints warning messages for extraction failures
        - Catches and logs exceptions during parsing
    """
    try:
        # Clear the global dictionary
        global TARGET_FW_VERSIONS
        TARGET_FW_VERSIONS.clear()

        # Parse info-v0
        info_file_path = os.path.join(data_temp_dir, "dump-info-v0")
        if os.path.exists(info_file_path):
            try:
                with open(info_file_path, 'r') as f:
                    fw_data = json.load(f)

                # Extract versions from the Members array
                if 'Members' in fw_data:
                    for component in fw_data['Members']:
                        name = component.get('Name', '')
                        version = component.get('Version', '')
                        if name and version:
                            TARGET_FW_VERSIONS[name] = version
                        # Add BMC and CEC per/postfixs - pending remove once added to bundle.
                        if "BMC_FW" in name:
                            TARGET_FW_VERSIONS[name] = 'BF-'+version
                        if name == "BF3_CEC_FW":
                            TARGET_FW_VERSIONS[name] = version+'_n02'
            except json.JSONDecodeError as e:
                print(f"Error parsing JSON in info-v0: {e}")
            except Exception as e:
                print(f"Error processing info-v0 file: {e}")

        # Parse bl33-v0 for UEFI version
        bl33_file_path = os.path.join(data_temp_dir, "dump-bl33-v0")
        if os.path.exists(bl33_file_path):
            try:
                decompress_tool = None

                check_zcat_cmd = "which zcat"
                _, zcat_rc = run_command(check_zcat_cmd)

                if zcat_rc == 0:
                    decompress_tool = "zcat"
                else:
                    check_gunzip_cmd = "which gunzip"
                    _, gunzip_rc = run_command(check_gunzip_cmd)

                    if gunzip_rc == 0:
                        decompress_tool = "gunzip -c"
                    else:
                        print("Warning: Neither zcat nor gunzip is available for UEFI version extraction")

                # Proceed only if we have a decompression tool
                if decompress_tool:
                    cmd = f"{decompress_tool} {bl33_file_path} | strings -el | grep 'BlueField' | cut -d':' -f 2"
                    uefi_version, rc = run_command(cmd)
                    if rc == 0 and uefi_version.strip():
                        TARGET_FW_VERSIONS["SHARED_UEFI"] = uefi_version.strip()
                    else:
                        print("Warning: Failed to extract UEFI version from bl33-v0.")
            except Exception as e:
                print(f"Error extracting UEFI version: {e}")

        # Parse bl2-v0 for ATF version
        bl2_file_path = os.path.join(data_temp_dir, "dump-bl2-v0")
        if os.path.exists(bl2_file_path):
            try:
                cmd = f"strings {bl2_file_path} | grep '(release)' | cut -d':' -f 2"
                atf_version, rc = run_command(cmd)
                if rc == 0 and atf_version.strip():
                    TARGET_FW_VERSIONS["SHARED_ATF"] = atf_version.strip()
                else:
                    print("Warning: Failed to extract ATF version from bl2-v0.")
            except Exception as e:
                print(f"Error extracting ATF version: {e}")

    except Exception as e:
        print(f"Error parsing firmware info: {e}")


def display_target_image_firmware_versions():
    """
    Display the firmware versions stored in the global dictionary.
    Shows components in a table with BlueField-2 and BlueField-3 versions side by side.
    Shared components are displayed for both device types.
    """
    if not TARGET_FW_VERSIONS:
        print("No firmware version information available.")
        return
    # Organize data by component type
    components = {}
    for name, version in TARGET_FW_VERSIONS.items():
        if '_' not in name:
            # Treat as a shared component
            component_type = name
            components[component_type] = {"BlueField-2": version, "BlueField-3": version}
            continue

        # Handle shared components (no BF2_/BF3_ prefix)
        prefix, component_type = name.split('_', 1)
        if name.startswith("SHARED_"):
            bf_type = "SHARED"
        # Handle BlueField-specific components (with BF2_/BF3_ prefix)
        elif name.startswith("BF2_") or name.startswith("BF3_"):
            bf_type = "BlueField-2" if prefix == "BF2" else "BlueField-3"
        else:
            # Skip unknown format
            continue

        # Initialize component entry if needed
        if component_type not in components:
            components[component_type] = {"BlueField-2": "N/A", "BlueField-3": "N/A"}

        # Set version based on device type
        if bf_type == "SHARED":
            # For shared components, set the same version for both BF2 and BF3
            components[component_type]["BlueField-2"] = version
            components[component_type]["BlueField-3"] = version
        else:
            # For device-specific components, set version only for that device type
            components[component_type][bf_type] = version

    # Prepare data for table
    headers = ["Component", "BlueField-2", "BlueField-3"]
    table = []

    # First handle DOCA if it exists
    if "DOCA" in components:
        versions = components["DOCA"]
        table.append([
            "DOCA",
            versions["BlueField-2"],
            versions["BlueField-3"]
        ])
        # Remove DOCA from the dictionary to avoid duplication
        del components["DOCA"]

    # Then add all remaining components in their original order
    for component_type, versions in components.items():
        if "GI" in component_type and versions["BlueField-2"] == "N/A":
            versions["BlueField-2"] = "Not Supported"
        display_name = "Bootimage" if component_type == "ATF_UEFI" else component_type
        table.append([
            display_name,
            versions["BlueField-2"],
            versions["BlueField-3"]
        ])

    print("\nTarget Firmware Versions of the provided image:")
    print(tabulate(table, headers=headers, tablefmt="fancy_grid"))


def _format_value(value):
    """Convert None to "N/A" otherwise return the value as is."""
    return value if value is not None else "N/A"


def _is_version_match(target_value, actual_value):
    """Check if target version matches or is contained in actual version."""
    if target_value is None or actual_value is None:
        return False

    target_str = str(target_value)
    actual_str = str(actual_value)

    if target_str == "N/A" or actual_str == "N/A":
        return False

    return target_str in actual_str or actual_str in target_str


def _analyze_device_versions(bf_device):
    """
    Analyze versions for a device and return component comparisons.

    Args:
        bf_device (BFDevice): The device to analyze

    Returns:
        dict: Contains target and post versions, match status, and related info
    """
    bf_type = "BF3_" if bf_device.bf_type == "bf3" else "BF2_"

    # Get target versions based on BF type
    target_atf = _format_value(TARGET_FW_VERSIONS.get("SHARED_ATF", "N/A"))
    target_uefi = _format_value(TARGET_FW_VERSIONS.get("SHARED_UEFI", "N/A"))
    target_nic_fw = _format_value(TARGET_FW_VERSIONS.get(f"{bf_type}NIC_FW"))
    target_bmc_fw = _format_value(TARGET_FW_VERSIONS.get(f"{bf_type}BMC_FW"))

    # Get pre/post install values
    pre_atf = _format_value(bf_device.pre_install_atf_fw_version)
    pre_uefi = _format_value(bf_device.pre_install_uefi_fw_version)
    pre_nic = _format_value(bf_device.pre_install_nic_fw_version)
    pre_bmc = _format_value(bf_device.pre_install_bmc_fw_version)

    post_atf = _format_value(bf_device.post_install_atf_fw_version)
    post_uefi = _format_value(bf_device.post_install_uefi_fw_version)
    post_nic = _format_value(bf_device.post_install_nic_fw_version)
    post_bmc = _format_value(bf_device.post_install_bmc_fw_version)

    # Check match status for each component
    atf_match = _is_version_match(target_atf, post_atf)
    uefi_match = _is_version_match(target_uefi, post_uefi)
    nic_match = _is_version_match(target_nic_fw, post_nic)
    bmc_match = _is_version_match(target_bmc_fw, post_bmc)

    # Identify failed components
    failed_components = []
    if not atf_match:
        failed_components.append("ATF")
    if not uefi_match:
        failed_components.append("UEFI")
    if not nic_match:
        failed_components.append("NIC_FW")
    if not bmc_match:
        failed_components.append("BMC_FW")

    return {
        "bf_device": bf_device,
        "components": {
            "ATF": {
                "pre": pre_atf,
                "target": target_atf,
                "post": post_atf,
                "match": atf_match
            },
            "UEFI": {
                "pre": pre_uefi,
                "target": target_uefi,
                "post": post_uefi,
                "match": uefi_match
            },
            "NIC_FW": {
                "pre": pre_nic,
                "target": target_nic_fw,
                "post": post_nic,
                "match": nic_match
            },
            "BMC_FW": {
                "pre": pre_bmc,
                "target": target_bmc_fw,
                "post": post_bmc,
                "match": bmc_match
            }
        },
        "failed_components": failed_components,
        "upgrade_successful": len(failed_components) == 0
    }


def display_verification_results(devices_analysis):
    """
    Display verification results for each device showing pre, target, and post upgrade firmware versions.

    Args:
        devices_analysis (list): List of analyzed device data
    """
    # Define color function locally since it's only used here
    def _color_if_match(target_value, actual_value):
        """Apply green background to text if versions match."""
        # Define ANSI color codes
        GREEN_BG = "\033[42m"  # Green background
        RESET = "\033[0m"      # Reset formatting

        if _is_version_match(target_value, actual_value):
            return f"{GREEN_BG}{actual_value}{RESET}"
        return actual_value
    
    print("\nPost-installation verification results:")

    for i, analysis in enumerate(devices_analysis):
        if i > 0:
            print("\n" + "-" * 80 + "\n")  # Separator between devices

        bf_device = analysis["bf_device"]
        print(f"\nVerification for device: {bf_device.bf_pci_address} (PSID: {_format_value(bf_device.psid)})")

        table_data = []
        for component_name, component_data in analysis["components"].items():
            colored_post = _color_if_match(component_data["target"], component_data["post"])
            table_data.append([
                component_name,
                component_data["pre"],
                component_data["target"],
                colored_post
            ])

        # Display the table
        headers = ["Component", "Pre Upgrade", "Target", "Post Upgrade"]
        print(tabulate(table_data, headers=headers, tablefmt="fancy_grid"))


def display_verification_summary(devices_analysis, failed_devices: Optional[List[BFDevice]] = None):
    """
    Display a summary report of devices that passed and failed the upgrade process.
    For failed devices, it lists which components did not match the target versions.

    Args:
        devices_analysis (list): List of analyzed device data
        failed_devices (list): Optional list of devices that failed during installation
    """
    # Separate successful and failed devices
    successful_devices = []
    failed_devices_version_mismatch = []

    for analysis in devices_analysis:
        if analysis["upgrade_successful"]:
            successful_devices.append(analysis["bf_device"])
        else:
            failed_devices_version_mismatch.append((analysis["bf_device"], analysis["failed_components"]))

    print("\n" + "="*80)
    print("UPGRADE SUMMARY REPORT")
    print("="*80)

    print("\nThe following devices passed upgrade successfully:")
    if successful_devices:
        for i, device in enumerate(successful_devices, 1):
            print(f"      {i}. {device.bf_pci_address} (PSID: {_format_value(device.psid)})")
    else:
        print("      None")

    print("\nThe following devices did not pass upgrades successfully:")

    # First show devices that completed installation but have version mismatches
    if failed_devices_version_mismatch:
        for i, (device, components) in enumerate(failed_devices_version_mismatch, 1):
            component_list = ", ".join(components)
            print(f"      {i}. {device.bf_pci_address} (PSID: {_format_value(device.psid)}) - "
                  f"{component_list} not matching to target image content")

    # Then show devices that failed during installation (if provided)
    start_index = len(failed_devices_version_mismatch) + 1
    if failed_devices:
        for i, device in enumerate(failed_devices, start_index):
            print(f"      {i}. {device.bf_pci_address} (PSID: {_format_value(device.psid)}) - "
                  f"Installation error")

    if not failed_devices_version_mismatch and not failed_devices:
        print("      None")

    print("\n" + "="*80)


def verify_post_installation_versions(bf_devices: List[BFDevice]) -> None:
    """
    Fetch and set post-installation firmware versions for devices in parallel,
    then display results with a summary report of successful and failed upgrades.

    Args:
        bf_devices (List[BFDevice]): All devices that were attempted for upgrade
    """
    print("\nValidating new FW on DPUs...")

    # Separate ready and failed devices
    # - ready_devices: completed installation, need version verification
    # - failed_devices: installation failed, will be reported as "Installation error"
    ready_devices = [device for device in bf_devices if device.dpu_ready]
    failed_devices = [device for device in bf_devices if not device.dpu_ready]

    # Create threads to run verification in parallel - only for ready devices
    threads = []
    for bf_device in ready_devices:
        thread = threading.Thread(
            target=bf_device._update_post_installation_fw_versions,
            name=f"VerifyThread-{bf_device.rshim_device}"
        )
        thread.start()
        threads.append(thread)

    # Wait for all threads to complete
    for thread in threads:
        thread.join()

    # Only analyze ready devices
    devices_analysis = [_analyze_device_versions(device) for device in ready_devices]

    # Display results only for ready devices
    if devices_analysis:
        display_verification_results(devices_analysis)

    # Display summary including both ready and failed devices
    display_verification_summary(devices_analysis, failed_devices)


def _check_bmc_reboot_in_config(config_file_path: str) -> bool:
    """
    Check if BMC_REBOOT=yes or equivalent is present in a configuration file.

    Args:
        config_file_path (str): Path to the configuration file to check

    Returns:
        bool: True if BMC_REBOOT is set to yes/true/1, False otherwise
    """
    if not config_file_path or not os.path.exists(config_file_path):
        return False

    try:
        with open(config_file_path, "r") as f:
            config_content = f.read()

        # First look for a match with actual 'yes'/'true'/'1' values
        bmc_reboot_match = re.search(r'BMC_REBOOT\s*=\s*["\']?(yes|true|1)["\']?', config_content, re.IGNORECASE)
        if bmc_reboot_match:
            return True

        return False
    except Exception as e:
        print(f"WARNING: Error checking BMC_REBOOT in config file: {e}")
        return False


def check_bmc_reboot_configured(bf_devices: List[BFDevice]) -> bool:
    """
    Check if BMC_REBOOT is configured for all devices that had their BMC firmware updated
    to the target version.

    Args:
        bf_devices (List[BFDevice]): List of BlueField devices to check

    Returns:
        bool: True if BMC_REBOOT is configured for all devices that need it, False otherwise
    """
    for device in bf_devices:
        # Skip devices that didn't have BMC firmware updated
        if not device.post_install_bmc_fw_version or device.post_install_bmc_fw_version == "N/A":
            continue

        # Get the target BMC firmware version for this device type
        bf_type_prefix = "BF3_" if device.bf_type == "bf3" else "BF2_"
        target_bmc_fw = TARGET_FW_VERSIONS.get(f"{bf_type_prefix}BMC_FW")

        # Only check for BMC_REBOOT if the installed version matches the target version
        if _is_version_match(target_bmc_fw, device.post_install_bmc_fw_version):
            # If BMC version matches the target, check for BMC_REBOOT configuration
            if not device.device_config_file or not _check_bmc_reboot_in_config(device.device_config_file):
                return False

    return True


def filter_excluded_devices(
    bf_devices: List[BFDevice], excluded_rshim_devices: List[str] = []
) -> Tuple[bool, List[BFDevice]]:
    """
    Filter out RSHIM devices that should be excluded from installation.
    
    Args:
        bf_devices: List of BFDevice objects to filter
        excluded_rshim_devices: List of RSHIM device names to exclude
        
    Returns:
        Tuple of (bool, List[BFDevice]) where:
            - bool: True if there are devices left after filtering, False otherwise
            - List[BFDevice]: List of devices after filtering out excluded ones

    Note:
        This function should be called after check_bf_rshim_devices, when rshim_device properties are set.
    
    """
    if not excluded_rshim_devices or not bf_devices:
        return True, bf_devices
    
    excluded_devices = set(excluded_rshim_devices)
    original_count = len(bf_devices)
    
    # Filter out the excluded devices
    filtered_devices = [
        device for device in bf_devices 
        if device.rshim_device not in excluded_devices
    ]
    
    # Check if any devices were excluded
    excluded_count = original_count - len(filtered_devices)
    if excluded_count > 0:
        print(f"\nINFO: Excluded {excluded_count} device(s) from installation as requested.")
    
    # Check if all devices were excluded
    if not filtered_devices:
        print(
            "\nSummary:\n"
            "WARNING: All available devices have been excluded from installation.\n"
            "No devices left to upgrade. Exiting.\n"
        )
        return False, []
    
    return True, filtered_devices

def display_firmware_comparison(bf_devices: List[BFDevice]) -> None:
    """
    Display a comparison table for each device showing current firmware versions
    versus the versions in the provided image bundle.

    Args:
        bf_devices (List[BFDevice]): List of BlueField devices to compare
    """
    if not TARGET_FW_VERSIONS:
        print("ERROR: No target firmware version information available from the image bundle.")
        return

    print("\nFirmware Version Comparison:")
    print("=" * 80)

    for i, device in enumerate(bf_devices):
        if i > 0:
            print("\n" + "-" * 80 + "\n")  # Separator between devices

        # Determine device type prefix for target versions
        bf_type_prefix = "BF3_" if device.bf_type == "bf3" else "BF2_"

        # Get target versions based on device type
        target_atf = _format_value(TARGET_FW_VERSIONS.get("SHARED_ATF", "N/A"))
        target_uefi = _format_value(TARGET_FW_VERSIONS.get("SHARED_UEFI", "N/A"))
        target_nic_fw = _format_value(TARGET_FW_VERSIONS.get(f"{bf_type_prefix}NIC_FW", "N/A"))
        target_bmc_fw = _format_value(TARGET_FW_VERSIONS.get(f"{bf_type_prefix}BMC_FW", "N/A"))
        target_cec_fw = _format_value(TARGET_FW_VERSIONS.get(f"{bf_type_prefix}CEC_FW", "N/A"))

        # Get current device versions
        current_atf = _format_value(device.pre_install_atf_fw_version)
        current_uefi = _format_value(device.pre_install_uefi_fw_version) 
        current_nic_fw = _format_value(device.pre_install_nic_fw_version)
        current_bmc_fw = _format_value(device.pre_install_bmc_fw_version)
        current_cec_fw = _format_value(device.pre_install_cec_fw_version)

        # For devices without pre_install versions set, need to get them
        if current_atf == "N/A" or current_uefi == "N/A":
            atf_version, uefi_version = get_atf_uefi_versions(device)
            current_atf = _format_value(atf_version)
            current_uefi = _format_value(uefi_version)

        print(f"Device: {device.bf_pci_address} (PSID: {_format_value(device.psid)}, Type: {device.bf_type.upper()})")

        # Prepare table data
        table_data = [
            ["ATF", current_atf, target_atf],
            ["UEFI", current_uefi, target_uefi],
            ["NIC_FW", current_nic_fw, target_nic_fw],
            ["BMC_FW", current_bmc_fw, target_bmc_fw],
            ["CEC_FW", current_cec_fw, target_cec_fw]
        ]

        # Display the comparison table
        headers = ["Component", "Current Version", "Image Version"]
        print(tabulate(table_data, headers=headers, tablefmt="fancy_grid"))

    print("\n" + "=" * 80)
